/**
 *  \file msglog.c
 *  \brief Functions for printing messages, warnings and errors.
 *
 *  This module provides output printing facilities.
 * 
 * include LICENSE
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include <msglog.h>
#include <duprintf.h>
 
#ifdef TRACE_MEM
#include <tracemem.h>
#endif
        

static FILE *msg_fd;
static char *msg_progname ;
static int msg_flags;
static int msg_verbose_level;
static int msg_dbg_msk;
static char *msg_filename;
static MsgFunc msg_displayFunc;

#define MSG_FMT_DATE \
"%2d %3s %4d  %2d:%02d:%02d",  \
ptm->tm_mday, msg_month_of_year[ptm->tm_mon], (1900 + ptm->tm_year), \
ptm->tm_hour, ptm->tm_min, ptm->tm_sec

const char *msg_month_of_year[] = {
    "Jan",  "Feb",  "Mar",  "Apr", "May", "Jun",
    "Jul",  "Aug",  "Sep",  "Oct", "Nov", "Dec",
};

const char *msg_day_of_week[] = {
        "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

char *msg_typename[] = {
   "DEBUG",
      "INFO",
      "WARNING",
      "ERROR",
      "FATAL",
};

/* see man console_codes */
#define COL(x)  "\033[" #x ";1m"
#define RED     COL(31)
#define GREEN   COL(32)
#define YELLOW  COL(33)
#define BLUE    COL(34)
#define MAGENTA COL(35)
#define CYAN    COL(36)
#define WHITE   COL(37)
#define GRAY    "\033[0m"       /* end of color */

char *msg_color[] = {
   "",          /* Debug */
      "",         /* info  */
      GREEN,      /* Warning */
      RED,        /* Error */
      BLUE,       /* Fatal */
};

/*
 * if given filename store message in file
 * if given displayFunc call func to display messge
 *   except for debug message
 */
void msg_set_func( MsgFunc displayFunc )
{
   msg_displayFunc = displayFunc;
   if ( displayFunc ) {
     msg_flags &= ~ MSG_F_TYPE;
   }
}

void msg_initlog(char *name, int flags, char *filename,	 MsgFunc displayFunc )
{
   msg_progname = name;
   msg_flags = flags;
   msg_verbose_level = 1;
   msg_dbg_msk = 1;
   msg_set_func( displayFunc );
   if ( filename ) {
      msg_openlog(filename, "w");
   }
}

void msg_exit(void)
{
   if ( msg_flags & MSG_F_SIGNAL) {
      kill(getpid(), SIGQUIT );
   }
   exit(1);
}

void msg_set_flags(int flags)
{
   msg_flags = flags;
}
void msg_set_level(int level)
{
   msg_verbose_level = level;
}
int msg_get_level(void)
{
   return msg_verbose_level;
}
void msg_set_dbg_msk(int msk)
{
   msg_dbg_msk = msk;
}

/*
 * msg_openlog :
 * send log messages to a file if needed
 */

FILE *msg_openlog(char *filename, char *mode)
{
   msg_filename = filename;
   if ((msg_fd = fopen(filename, mode)) == NULL ) {
      msg_fatal(_("Can't open file '%s': %s"), filename,  strerror (errno));
   }
   atexit(msg_closelog);
   return msg_fd;
}

void msg_closelog(void)
{
   if ( msg_fd != stderr ) {
      fclose(msg_fd);
   }
}

void msg_str_time_date( char *strdate )
{
   time_t now;
   struct tm *ptm;
   
   time(&now) ;
   ptm = localtime(&now);
   sprintf(strdate, MSG_FMT_DATE );
}

void private_message( int level, int msgtype, const char *func, char *file,
		      int line, char *fmt, ... )
{
   va_list ap;
   char date[128];
   char text[128];
   char *ffmt;
   char *p;
   char *color_s = "";
   char *color_e = "";

   if ( msgtype == MSG_T_DEBUG ){
      if ( (msg_dbg_msk & level) == 0 ){ 
	 return ;
      }
   } else if ( level > msg_verbose_level ) {
      return ;
   }
   date[0] = 0;
   if ( msg_flags & MSG_F_DATE ) {
      msg_str_time_date(date);
      strcat(date, ": ");
   }
   if ( msg_flags & MSG_F_PROG ) {
      if ( msg_progname && *msg_progname ) {
	 p = date + strlen(date);
	 snprintf(p, sizeof(date) - strlen(date), "%s: ", msg_progname);
      }
   }
   
   text[0] = 0;
   if ( msgtype == MSG_T_DEBUG || msg_flags & MSG_F_TYPE ) {
      p = text + strlen(text);
      snprintf(p, sizeof(text) - strlen(text), "%s: ", msg_typename[msgtype]);
   }
   if ( msg_flags & MSG_F_FUNC ) {
      p = text + strlen(text);
      snprintf(p, sizeof(text) - strlen(text), "%s: ", func);
   }
   if ( msg_flags & MSG_F_FILE ) {
      p = text + strlen(text);
      snprintf(p, sizeof(text) - strlen(text), "%s:%d: ", file, line );
   }
   if ( msg_flags & MSG_F_COLOR ) {
      color_s = msg_color[msgtype];
      if ( msgtype > MSG_T_INFO ){
	 color_e = GRAY;
      }
   }

   va_start(ap, fmt);
   ffmt = app_strdup_vprintf (fmt, ap);
   va_end(ap);
   
   if ( msg_flags & MSG_F_FOLD ) {
      if ( strlen(date) + strlen(text) + strlen(ffmt) > 80 ){
	 strcat(text, "\n    ");
      }
   }
   
   if ( ! msg_fd ) {
      msg_fd = stderr;
   }
   if ( msgtype == MSG_T_DEBUG ){
      fprintf(stderr, "%s%s%s\n", date, text, ffmt);
   } else if ( msg_displayFunc ){
      char message[1024];
      sprintf(message, "%s%s%s\n", date, text, ffmt);
      (msg_displayFunc) ( msgtype, message );
   } else {
      fprintf(msg_fd, "%s%s%s%s%s\n", color_s, date, text, ffmt, color_e);
      fflush(msg_fd);
   }
   app_free(ffmt);
}
