/*
 * sockcon.c - socket type connection handling
 * 
 * include LICENSE
 */

#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <strmem.h>
#include <sockcon.h>
 
#ifdef TRACE_MEM
#include <tracemem.h>
#endif
        

#if 0
static int con_wait_ready(SockCon *p);
#endif
//static int con_init_sockaddr( char *host,  struct sockaddr_in *bind_addr);

static int con_init_sockaddr( SockCon *cnx);
static int con_write (SockCon *cnx, char *buf, int len, int flags);
static int con_read (SockCon *cnx, char *buf, int len, int flags);


/*
 *** \brief Allocates memory for a new SockCon object.
 *
 * con_new - connection init
 */
SockCon *con_new( char *host, int domain, int sockType, int proto, int port)
{
   SockCon *cnx;
   
   cnx = app_new0( SockCon, 1);
   con_construct( cnx, host, domain, sockType, proto, port);
   app_class_overload_destroy( (AppClass *) cnx, con_destroy );

   return cnx;
}
   
/** \brief Constructor for the SockCon object. */

void con_construct( SockCon *cnx, char *host, int domain, int sockType, int proto, int port)
{
   app_class_construct( (AppClass *) cnx );

   cnx->host_to_connect = host;
   cnx->domain = domain;
   cnx->sockType = sockType;
   cnx->port = port;
   cnx->type = CON_TEMP;
   cnx->wrfunc = con_write;
   cnx->rdfunc = con_read;
   int ret;

   if ((cnx->s = socket(domain, sockType, proto)) < 0) {
      msg_error(_("Can't open socket - err = %s"), strerror (errno) ) ;
      cnx->status = -1;
      return ;
   }
//   if ( con_init_sockaddr( host, &(cnx->bind_addr)) == 0 ) {
   if ( (ret = con_init_sockaddr( cnx)) == 0 ) {
      /* server init */
      cnx->bind_addr.sin_family = domain;
      cnx->bind_addr.sin_addr.s_addr = INADDR_ANY;
   } else if ( ret < 0) {
      cnx->status = -1;
      return ;
   }
   

   cnx->bind_addr.sin_port = htons(cnx->port);

   msg_dbgl(DBG_8, "init ok: sin_port %d: sin_family %d: port %d", 
	   cnx->bind_addr.sin_port, cnx->bind_addr.sin_family, cnx->port );

   cnx->recvbuf = dbuf_new(CON_RECVBUF_SIZE);
   return ;
}

static int con_init_sockaddr( SockCon *cnx)
{
   struct in_addr in_ad;
   struct hostent *hp;

//   memset(bind_addr, 0, sizeof(struct sockaddr_in) );
   if ( cnx->host_to_connect && cnx->host_to_connect[0] ){
      /* client connection */
      if (inet_aton(cnx->host_to_connect, &in_ad) == 0 ) {
	 /* ad is not an ip address - it is a name, so try to get the addres */
	 hp = gethostbyname( cnx->host_to_connect );
	 if (hp == NULL) {
	    msg_error(_("host address information for %s not found"), cnx->host_to_connect );
	    return -1;
	 }
      } else {
	 hp = gethostbyaddr(&in_ad, sizeof (struct in_addr), AF_INET);
	 if (hp == NULL) {
	    msg_warningl( 2, _("host name information for %s not found"), cnx->host_to_connect );
	    cnx->full_name_host = app_strdup("No name found");
	    memcpy((char *) &(cnx->bind_addr.sin_addr), (char *) &in_ad, sizeof(struct in_addr ) );
	    cnx->bind_addr.sin_family = AF_INET ;
	    return 1;
	 }
      }
      cnx->full_name_host = app_strdup(hp->h_name);
      memcpy((char *) &(cnx->bind_addr.sin_addr), (char *) *hp->h_addr_list, hp->h_length );
      cnx->bind_addr.sin_family = hp->h_addrtype ;
      return 1;
   }
   return 0;
}

/** \brief Destructor for the SockCon object. */

void con_destroy(void *cnx)
{
   SockCon *this = (SockCon *) cnx;

   if (cnx == NULL) {
      return;
   }
   con_set_blocking(this, CON_BLOCKING);
   if ( this->closefunc ) {
      this->closefunc ( cnx, 1 );
   }
   shutdown( this->s, 2);
   if ( close(this->s) != 0 ){ 
      msg_warning(_("Can't close socket %d - %s"), this->s, strerror(errno) ) ;
   }

   app_free(this->full_name_host);
   app_free(this->connected_ip);
   app_free(this->connected_to);
   
   dbuf_destroy(this->recvbuf);
   app_class_destroy( cnx );
}


/*
 * Set non blocking i/o.
 */
 
void con_set_blocking(SockCon *p, int blocking)
{
   p->blocking = blocking &  CON_BLOCKING_MSK;
   if ( p->blocking & CON_BLOCKING ) {
      int ret = fcntl(p->s, F_GETFL ) ;
      if ( ret & O_NDELAY ) {
	 fcntl(p->s, F_SETFL, (ret & ~O_NDELAY) ) ;
      }
   } else {
      fcntl(p->s, F_SETFL, O_NDELAY ) ;
   }
}

void con_set_timeout(SockCon *cnx, int timeout)
{
   cnx->timeout = timeout;
}

/*
 * update usefull info about peers
 */
static void con_update(SockCon *cnx)
{
   struct sockaddr_in name;
   socklen_t l = sizeof( struct sockaddr );
   
   if ( cnx->type & CON_TEMP ) {
      return;
   }
   if ( getpeername( cnx->s, (struct sockaddr *) &name, &l ) == 0 ) {
      app_free(cnx->connected_ip);
      app_free(cnx->connected_to);
      cnx->connected_ip = app_strdup(inet_ntoa(name.sin_addr));
      cnx->connected_to = con_get_host_name(cnx->connected_ip);
      msg_dbgl(DBG_8, "Connected: %s %s", cnx->connected_to, cnx->connected_ip);
   } else {
      msg_warning(_("Can't get peer name.- %s"), strerror(errno) );
   }
}

SockCon *con_accept(SockCon *pser, int checkFlags)
{
   int s ;
   SockCon *cnx;
   socklen_t l = sizeof(pser->bind_addr) ;
 
   if ( (checkFlags & CON_CHECK_READ) && ! con_is_ready(pser, checkFlags ) ){
      return NULL;
   }
   s = accept(pser->s, (struct sockaddr *) &pser->bind_addr, &l )  ;
   if ( s > 0 ) {
      cnx = app_new0(SockCon, 1);
      memcpy(cnx, pser, sizeof(SockCon));
      cnx->s = s;
      con_set_blocking( cnx, checkFlags);
      cnx->type = CON_SERVER;
      con_is_ready(cnx, CON_CHECK_WRITE | CON_CHECK_UPDATE );
      app_class_ref( (AppClass *) cnx );
      cnx->recvbuf = dbuf_new(CON_RECVBUF_SIZE);

      msg_dbgl(DBG_8, "Accepted: %s %s", cnx->connected_to, cnx->connected_ip);
      return(cnx) ;
   } else {
     msg_error( _("Can't accept socket - %s"), strerror(errno) ) ;
   }
      
   return(NULL) ;
}
/*
 * get connection from inetd
 */
SockCon *con_acquire(int flags)
{
   SockCon *cnx;
   
   cnx = app_new0(SockCon, 1);
   cnx->s = dup(0); /* 0 in daemon mode */;
   close(0);
   con_set_blocking( cnx, flags);
   cnx->type = CON_SERVER;
   
   cnx->recvbuf = dbuf_new(CON_RECVBUF_SIZE);
   con_is_ready(cnx, CON_CHECK_WRITE | CON_CHECK_UPDATE);

   msg_dbgl(DBG_8, "Accepted: %s %s", cnx->connected_to, cnx->connected_ip);
   return cnx;
}


/*
 * client connection open
 */

int con_connect (SockCon *pcli)
{
   if ( pcli->timeout ) {
      alarm(pcli->timeout);
   }
   if ( connect(pcli->s, (struct sockaddr *) &pcli->bind_addr, sizeof(pcli->bind_addr)) < 0) {
      if ( errno != EINPROGRESS ) {
	 msg_warning(_("Can't connect socket - err = %d %s\n"), errno, strerror(errno) ) ;
	 return(-1) ;
      }
   }
   alarm(0);
   pcli->type = CON_CLIENT;
   
   if ( ! con_is_ready(pcli, CON_CHECK_WRITE | CON_CHECK_UPDATE ) ) {
      return -1;
   }
      
   return(pcli->s) ;
}

int con_listen(SockCon *pser, int backlog )
{
   if ( bind(pser->s, (struct sockaddr *) &pser->bind_addr, sizeof(pser->bind_addr)) < 0 &&
       errno != EADDRINUSE ) {
      msg_error( _("Can't bind socket - %s"),  strerror(errno) ) ;
      return(-1) ;
   }
   if ( listen(pser->s, backlog)  < 0 ) {
      msg_error( _("Can't listen socket - %s"), strerror(errno) ) ;
      return(-1) ;
   }
   return(pser->s) ;
}


/*
 * Check if connection is ready (ready to read, write)
 */
int con_is_ready(SockCon *p, int check)
{   
   int i;
   fd_set fdset ;
   int nfds;
   struct timeval sel_timeout ;

   sel_timeout.tv_sec = 0 ;
   if ( check & CON_CHECK_TIMEOUT) {
      sel_timeout.tv_sec = p->timeout ;
   }
   sel_timeout.tv_usec = 0 ;

      
   FD_ZERO(&fdset) ;
   FD_SET(p->s, &fdset) ;
   nfds = p->s + 1;

   if ( check & CON_CHECK_READ) {
      i = select(nfds, &fdset, NULL, NULL, &sel_timeout ) ;
   } else {
      i = select(nfds, NULL, &fdset, NULL, &sel_timeout ) ;
   }
   if ( i == -1 ) {
      msg_error(_("select error '%s'"), strerror(errno) ) ;
   } else if ( i > 0 ) {
      if ( (check & CON_CHECK_WRITE) && p->connected_ip == NULL ) { 
	 con_update(p);
      }
      return 1;
   }
   return 0;
}

#if 0
static int con_wait_ready(SockCon *p)
{   
   int i;

   for( i = 0; i < 100 ; i++) {
      if ( con_is_ready(p, CON_CHECK_WRITE) ){
	 return 0;
      }
      usleep(100000);
   }
   msg_error(_("Timeout: Connect socket not ready")) ;
//   close(s) ;
   return -1;
}
#endif
		
void con_SetLinger(int sockfd)
{
   struct linger li;
 
   li.l_onoff = 1;
   li.l_linger = 120;      /* 2 minutes, but system ignores field. */

   /*
    * Have the system make an effort to deliver any unsent data,
    * even after we close the connection.
    */
   if (setsockopt(sockfd, SOL_SOCKET, SO_LINGER, (char *) &li, (int) sizeof(li)) < 0)
      msg_warning(_("Linger mode could not be enabled."));
}       /* SetLinger */


void con_SetKeepalive(int sockfd)
{
   int on = 1;

   if (setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE, (char *) &on, (int) sizeof(on)) < 0)
      msg_warning(_("Keepalive mode could not be enabled."));
}       /* SetKeepalive */

void con_SetPriority(int sockfd)
{
/* Mac OS X doesn't implement SO_PRIORITY.
 * Maybe this should be rather handled on the autoconf level with a test program and a macro in config.h.
 */
#ifdef SO_PRIORITY
   int on = 1;

   if (setsockopt(sockfd, SOL_SOCKET, SO_PRIORITY, (char *) &on, (int) sizeof(on)) < 0)
      msg_warning(_("Priority mode could not be enabled."));
#else /* SO_PRIORITY */
   msg_warning(_("Priority mode not supported by the OS."));
#endif /* SO_PRIORITY */
}       /* SetKeepalive */

void con_SetSndbuf(int sockfd, int size)
{
   int on ;

   on = size;
   if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, (char *) &on, (int) sizeof(on)) < 0)
     msg_warning(_("SetSndbuf mode could not be enabled."));
}       /* SetSndbuf */




int con_set_port(char *service, char *proto, int *port)
{
   struct servent *sp ;
   
   if ( service && service[0] ) {
      if ((sp = getservbyname(service, proto)) == NULL) {
	 if ( *port == 0 ) {
	    msg_error( _("Unkown service '%s'- port %d"), service, *port) ;
	    return(-1) ;
	 }
      } else {
	 *port = (unsigned int) ntohs(sp->s_port);
      }
      return *port;
   }
   return -1;
}

/*
 * full qualified name
 * If name is null try to return local host
 */
char *con_get_host_name(char *name)
{
   struct hostent *hp;
   struct in_addr in_ad;
   char localHost[128];

   if ( name == NULL ) {
      gethostname(localHost, sizeof(localHost));
      name = localHost;
   }
   if ( name[0] == 0 ) {
       return NULL ;
   }
   if (inet_aton(name, &in_ad) == 0 ) {
      /* ad is not an ip address */
      hp = gethostbyname( name );
   } else {
      /* ad is an ip address */
      hp = gethostbyaddr(&in_ad, sizeof (struct in_addr), AF_INET);
   }
   if (hp == NULL) {
      msg_errorl(2, _("host information from name %s not found"), name );
      return NULL;
   }
   return app_strdup(hp->h_name);
}

char *con_get_host_addr(char *name)
{
   struct hostent *hp;
   struct in_addr in_ad;

   if ( name == NULL || name[0] == 0 ) {
       return NULL ;
   }
   if (inet_aton(name, &in_ad) != 0 ) {
      /* ad is an ip address */
      return app_strdup(name) ;
   }

   hp = gethostbyname(name);
   if (hp == NULL) {
      msg_error( _("host information for '%s' not found"), name );
      return(NULL);
   }
   memcpy(&in_ad.s_addr, *(hp->h_addr_list), sizeof (struct in_addr));
   return app_strdup(inet_ntoa(in_ad));
}

static int con_write (SockCon *cnx, char *buf, int len, int flags)
{
   int ret;

   msg_infol( 5, _("Sending: '%s' len %d"), buf, len);

   ret = send( cnx->s, buf, len, flags );
   if ( ret < 0 ){
      msg_errorl( 2, _("Can't send - %s"), strerror(errno) );
   }
   return ret;
}

/*
 * send over the connection
 */

int con_send (SockCon *cnx, char *buf, int len, int flags)
{
   if ( len == 0 ) {
      len = strlen(buf);
   }
    return cnx->wrfunc ( cnx, buf, len, flags );
}

int con_dbuf_send (SockCon *cnx, DBuf *dbuf )
{
   int size = dbuf->len - dbuf->pos;
   int sent;

   sent = con_send (cnx, dbuf->s + dbuf->pos, size, 0);
   if ( sent > 0 ) {
      if ( sent < size ) {
         dbuf_set_pos( dbuf, dbuf->pos + sent );
      } else {
	 /* buffer completely sent */
	 dbuf_clear(dbuf);
      }
   }
   return sent;
}

/*
 * formatted send over the connection
 */

int con_fmt_send (SockCon *cnx, char *format, ...)
{
   va_list args;
   char buffer[BUFSIZ];
   
   va_start(args, format);
   vsnprintf(buffer, sizeof(buffer), format, args);
   va_end(args);

   return con_send( cnx, buffer, strlen(buffer), 0);
}

static int con_read (SockCon *cnx, char *buf, int len, int flags)
{
   int size;

   size = recv(cnx->s, buf, len, flags);

   if ( size <= 0 ){  
      if ( errno == EAGAIN ) {
	 return -2 ;
      }
      msg_errorl( 2, _("Receive error - %s"), strerror(errno) );
   }
   msg_dbgl( DBG_8, "Read: '%d' bytes", size);
   return size;
}

int con_recv (SockCon *cnx)
{
   int size;
   DBuf *dbuf = cnx->recvbuf; 

   if ( cnx->timeout ) {
      alarm(cnx->timeout);
   }
   size = cnx->rdfunc(cnx, dbuf->s + dbuf->len, dbuf->size - dbuf->len, 0);

   alarm(0);
   
   if ( size > 0 ){  
      dbuf_set_len( dbuf, dbuf->len + size  );
   }
   msg_dbgl( DBG_8, "Recv: '%d' bytes", size);
   return size;
}

int con_get_char(SockCon *cnx)
{
   DBuf *dbuf = cnx->recvbuf; 

   if (dbuf->pos >= dbuf->len) {
      dbuf_clear(dbuf);
      con_recv(cnx);
      if ( ! dbuf->len) {
         return -1;
      }
   }
   return (unsigned char) dbuf->s[dbuf->pos++];
}
